@@ -5,7 +5,7 @@ use strict;
 use warnings;
 use Log::Any '$log';
 
-our $VERSION = '0.10'; # VERSION
+our $VERSION = '0.11'; # VERSION
 
 use File::HomeDir;
 use HTTP::Daemon;
@@ -40,6 +40,7 @@ has access_log_path        => (is => 'rw');
 has http_ports             => (is => 'rw', default => sub{[]});
 has https_ports            => (is => 'rw', default => sub{[]});
 has unix_sockets           => (is => 'rw', default => sub{[]});
+has timeout                => (is => 'rw', default => sub{120});
 has require_root           => (is => 'rw', default => sub{0});
 has ssl_key_file           => (is => 'rw');
 has ssl_cert_file          => (is => 'rw');
@@ -48,7 +49,6 @@ has max_requests_per_child => (is => 'rw', default=>sub{1000});
 has _daemon                => (is => 'rw'); # SHARYANTO::Proc::Daemon::Prefork
 has _server_socks          => (is => 'rw'); # store server sockets
 has _app                   => (is => 'rw'); # store PSGI app
-has _client                => (is => 'rw'); # store client data
 has product_name           => (is => 'rw');
 has product_version        => (is => 'rw');
 
@@ -76,7 +76,7 @@ sub BUILD {
     }
     unless (defined $self->product_version) {
         no strict;
-        $self->product_version(${ref($self)."::VERSION"} // "0.0");
+        $self->product_version($Gepok::VERSION // "0.0");
     }
     unless ($self->_daemon) {
         my $daemon = SHARYANTO::Proc::Daemon::Prefork->new(
@@ -135,8 +135,12 @@ sub _after_init {
     $ary = $self->unix_sockets;
     if (defined($ary) && ref($ary) ne 'ARRAY') { $ary = [split /\s*,\s*/,$ary] }
     for my $path (@$ary) {
+        my %args;
+        $args{Reuse}   = 1;
+        $args{Timeout} = $self->timeout;
+        $args{Local}   = $path;
         $log->infof("Binding to Unix socket %s (http) ...", $path);
-        my $sock = HTTP::Daemon::UNIX->new(Local=>$path);
+        my $sock = HTTP::Daemon::UNIX->new(%args);
         die "Unable to bind to Unix socket $path" unless $sock;
         push @server_socks, $sock;
         push @server_sock_infos, "$path (unix)";
@@ -145,7 +149,9 @@ sub _after_init {
     $ary = $self->http_ports;
     if (defined($ary) && ref($ary) ne 'ARRAY') { $ary = [split /\s*,\s*/,$ary] }
     for my $port (@$ary) {
-        my %args = (Reuse => 1);
+        my %args;
+        $args{Reuse}   = 1;
+        $args{Timeout} = $self->timeout;
         if ($port =~ /^(?:0\.0\.0\.0)?:?(\d+)$/) {
             $args{LocalPort} = $1;
         } elsif ($port =~ /^(\d+\.\d+\.\d+\.\d+):(\d+)$/) {
@@ -165,9 +171,9 @@ sub _after_init {
     $ary = $self->https_ports;
     if (defined($ary) && ref($ary) ne 'ARRAY') { $ary = [split /\s*,\s*/,$ary] }
     for my $port (@$ary) {
-        my %args = (Reuse => 1);
-        # currently commented out, hangs with larger POST
-        #$args{Timeout} = 180;
+        my %args;
+        $args{Reuse}   = 1;
+        $args{Timeout} = $self->timeout; # can hang with larger POST?
 
         $args{SSL_key_file}  = $self->ssl_key_file;
         $args{SSL_cert_file} = $self->ssl_cert_file;
@@ -213,6 +219,8 @@ sub _main_loop {
         my @ready = $sel->can_read();
         for my $s (@ready) {
             my $sock = $s->accept();
+            # sock can be undef
+            next unless $sock;
             $self->{_connect_time} = [gettimeofday];
             $self->_set_label_serving($sock);
             $self->_daemon->update_scoreboard({
@@ -221,6 +229,8 @@ sub _main_loop {
                 state => "R",
             });
             while (my $req = $sock->get_request) {
+                $self->{_client_proto} =
+                    $sock->proto_ge("1.1") ? "HTTP/1.1" : "HTTP/1.0";
                 $self->{_finish_req_time} = [gettimeofday];
                 $self->_daemon->update_scoreboard({state => "W"});
                 my $res = $self->_handle_psgi($req, $sock);
@@ -238,96 +248,126 @@ sub _finalize_response {
 
     $self->{_sock_peerhost} = $sock->peerhost; # cache first before close
 
-    $self->_client({});
     if ($env->{'psgix.harakiri.commit'}) {
-        $self->_client->{keepalive} = 0;
-        $self->_client->{harakiri}  = 1;
+        $self->{_client_keepalive} = 0;
+        $self->{_client_harakiri}  = 1;
     }
 
-    my $protocol = $env->{SERVER_PROTOCOL};
-    my $status   = $res->[0];
-    my $message  = status_message($status);
+    my $server_proto = $env->{SERVER_PROTOCOL};
+    my $client_proto = $self->{_client_proto};
+    my $status       = $res->[0];
+    my $message      = status_message($status);
     $self->{_res_status} = $status;
 
     # generate HTTP status + response headers
 
     my(@headers, %headers);
-    push @headers, "$protocol $status $message";
+    push @headers, "$server_proto $status $message";
     push @headers, "Server: ".
             $self->product_name."/".$self->product_version;
     while (my ($k, $v) = splice @{$res->[1]}, 0, 2) {
-        next if $k eq 'Connection';
         push @headers, "$k: $v";
         $headers{lc $k} = $v;
     }
 
-    my $chunked;
-    if ($protocol eq 'HTTP/1.1') {
-        $chunked = 1;
-        $self->_client->{keepalive} //= 1;
-        if (my $te = $headers{'transfer-encoding'}) {
-            if ($te eq 'chunked') {
-                #$log->trace("Chunked transfer-encoding set for response");
-                $chunked = 1;
-            }
-        }
-    };
-    push @headers, "Transfer-Encoding: chunked" if $chunked;
-
-    if (!exists $headers{'content-length'}) {
-        #$log->trace("Disabling keep-alive after sending unknown length ".
-        #                "body on $protocol");
-        $self->_client->{keepalive} = 0;
-    }
-    if ($self->_client->{keepalive}) {
-        push @headers, 'Connection: keep-alive';
-    } else {
-        push @headers, 'Connection: close';
-    }
     if (!$headers{date}) {
         push @headers, "Date: " . time2str(time());
     }
 
-    # Buffer the headers so they are sent with the first write() call
-    # This reduces the number of TCP packets we are sending
-    syswrite $sock, join($CRLF, @headers, '') . $CRLF;
+    my $keepalive;
+    if ($env->{HTTP_CONNECTION}) {
+        $keepalive = $env->{HTTP_CONNECTION} =~ /alive/i ? 1:0;
+    }
+    # default is keep-alive for HTTP/1.1, but close for HTTP/1.0
+    $keepalive //= ($client_proto eq 'HTTP/1.1' ? 1:0);
+    # normally HTTP::Daemon prints this, but we're not sending response using
+    # HTTP::Daemon
+    push @headers, "Connection: ".($keepalive ? "Keep-Alive" : "Close");
 
-    my $body_size = 0;
+    my $chunked;
+    my $cl = $headers{'content-length'};
+    if ($client_proto eq 'HTTP/1.1') {
+        if ($status =~ /^[123]/ && (!defined($cl) || $cl)) {
+            $chunked = 1;
+        }
+        if (my $te = $headers{'transfer-encoding'}) {
+            $chunked = $te eq 'chunked';
+        }
+    } else {
+        # "A server MUST NOT send transfer-codings to an HTTP/1.0 client."
+        $chunked = 0;
+    }
+    push @headers, "Transfer-Encoding: chunked" if $chunked;
+    $self->{_chunked} = $chunked;
+
+    #warn "chunked=$chunked, keep-alive=$keepalive, client_proto=$client_proto";
+
+    if ($client_proto le 'HTTP/1.0' && $keepalive && !defined($cl)) {
+        # if HTTP/1.0 client requests keep-alive (like Wget), we need Content-Length
+        # so client knows when response ends.
+
+        # produce body first so we can calculate content-length
+        $self->_finalize_body($env, $res, $sock, 1);
+        push @headers, "Content-Length: ".$self->{_res_content_length};
+        syswrite $sock, join($CRLF, @headers, '') . $CRLF; # print header
+        syswrite $sock, $_ for @{$self->{_body}}; # print body
+    } else {
+        # print headers + body normally
+
+        syswrite $sock, join($CRLF, @headers, '') . $CRLF; # print header
+        $self->_finalize_body($env, $res, $sock);
+    }
+}
+
+# either print body to $sock, or store it in $self-> (for HTTP/1.0 Keep-Alive
+# clients)
+sub _finalize_body {
+    my ($self, $env, $res, $sock, $save) = @_;
+    my $cl = 0;
+    $self->{_body} = [] if $save;
     if (defined $res->[2]) {
         Plack::Util::foreach(
             $res->[2],
             sub {
                 my $buffer = $_[0];
                 my $len = length $buffer;
-                $body_size += $len;
-                if ($chunked) {
+                $cl += $len;
+                if ($self->{_chunked}) {
                     return unless $len;
                     $buffer = sprintf("%x", $len) . $CRLF . $buffer . $CRLF;
                 }
-                syswrite $sock, $buffer;
-                #$log->debug("Wrote " . length($buffer) . " bytes");
+                $self->_write_sock($sock, $save, $buffer);
             });
-        syswrite $sock, "0$CRLF$CRLF" if $chunked;
+        $self->_write_sock($sock, $save, "0$CRLF$CRLF") if $self->{_chunked};
     } else {
         return Plack::Util::inline_object(
             write => sub {
                 my $buffer = $_[0];
                 my $len = length $buffer;
-                $body_size += $len;
-                if ($chunked) {
+                $cl += $len;
+                if ($self->{_chunked}) {
                     return unless $len;
                     $buffer = sprintf("%x", $len) . $CRLF . $buffer . $CRLF;
                 }
-                syswrite $sock, $buffer;
-                #$log->trace("Wrote " . length($buffer) . " bytes");
+                $self->_write_sock($sock, $save, $buffer);
             },
             # poll_cb => sub { ... },
             close => sub {
-                syswrite $sock, "0$CRLF$CRLF" if $chunked;
+                $self->_write_sock($sock, $save, "0$CRLF$CRLF")
+                    if $self->{_chunked};
             }
         );
     }
-    $self->{_res_body_size} = $body_size;
+    $self->{_res_content_length} = $cl;
+}
+
+sub _write_sock {
+    my ($self, $sock, $save, $buffer) = @_;
+    if ($save) {
+        push @{$self->{_body}}, $buffer;
+    } else {
+        syswrite $sock, $buffer;
+    }
 }
 
 # run PSGI app, send PSGI response to client as HTTP response, and return it
@@ -396,6 +436,7 @@ sub _prepare_env {
         'gepok'                     => 1,
         'gepok.connect_time'        => $self->{_connect_time},
         'gepok.finish_request_time' => $self->{_finish_req_time},
+        'gepok.client_protocol'     => $self->{_client_proto},
     };
     $env->{HTTPS} = 'on' if $is_ssl;
 
@@ -412,8 +453,10 @@ sub _prepare_env {
 
 sub _set_label_serving {
     my ($self, $sock) = @_;
+    # sock can be undef when client timed out
+    return unless $sock;
 
-    my $is_unix = $sock && $sock->isa('HTTP::Daemon::UNIX');
+    my $is_unix = $sock->isa('HTTP::Daemon::UNIX');
 
     if ($is_unix) {
         my $sock_path = $sock->hostpath;
@@ -466,7 +509,7 @@ sub access_log {
         $req->method,
         __escape_quote($req->uri->as_string),
         $self->{_res_status},
-        $self->{_res_body_size} // "-",
+        $self->{_res_content_length} // "-",
         scalar($reqh->header("referer")) // "-",
         scalar($reqh->header("user-agent")) // "-",
     );
@@ -489,7 +532,7 @@ Gepok - PSGI server with built-in HTTPS support, Unix sockets, preforking
 
 =head1 VERSION
 
-version 0.10
+version 0.11
 
 =head1 SYNOPSIS
 
